###   Copyright 2004,   Xavier Neys   (neysx@gentoo.org)
# #
# #   This file is part of gorg.
# #
# #   gorg is free software; you can redistribute it and/or modify
# #   it under the terms of the GNU General Public License as published by
# #   the Free Software Foundation; either version 2 of the License, or
# #   (at your option) any later version.
# #
# #   gorg is distributed in the hope that it will be useful,
# #   but WITHOUT ANY WARRANTY; without even the implied warranty of
# #   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# #   GNU General Public License for more details.
# #
# #   You should have received a copy of the GNU General Public License
# #   along with Foobar; if not, write to the Free Software
###   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Run the stand-alone webserver and serve gentoo.org

require 'gorg/base'
require 'webrick'

class GentooServlet < WEBrick::HTTPServlet::FileHandler
  include Gorg
  
  def do_GET(req, res)
    hit = "#{$Config["root"]}#{req.path}"
    cacheName = req.path
    if FileTest.directory?(hit) then
      res.set_redirect(WEBrick::HTTPStatus::TemporaryRedirect, req.path.chomp("/") + "/index.xml") # Redirect dir to index.xml
    else
      if hit !~ /\.(xml)|(rdf)$/ then
        super # Use default FileHandler if not an xml file
      else
        if not FileTest.exist?(hit) then
          super # Use default FileHandler to handle 404 (file does not exist)
        else
          # Parse If-None-Match and If-Modified-Since request header fields if any
          ims=inm=nil
          begin
            ims = Time.parse(req['if-modified-since']) if req['if-modified-since']
            inm = split_header_etags(req['if-none-match']) if req['if-none-match']
          rescue
            # Just ignore ill-formated data
            nil
          end
          begin
            res['Charset'] = 'UTF-8'
            # Process xml file or return xml file if passthru=1
            if $Config['passthru'] && req.query && req.query["passthru"] && req.query["passthru"] != "0" then
              # passthru allowed by config and requested by visitor, return file as text/plain
              mstat = File.stat(hit)
              raise Gorg::Status::NotModified.new(mstat) if notModified?(mstat, inm, ims)
              res.body = IO.read(hit)
              res['Content-Type'] = 'text/plain'
              debug("Passthru granted for #{hit}")
            else
              res.body, mstat = Gorg::Cache.hit(cacheName, req.query, inm, ims)
              if res.body.nil? then
                # Cache miss, process file and cache result
                err, out, filelist = xproc(hit, req.query, true)
                # Cache the output only when 100% error-free, i.e. no warning 
                mstat = Gorg::Cache.store(out, cacheName, req.query, filelist) if err["xmlErrLevel"] == 0
                warn("#{err.collect{|e|e.join(':')}.join('; ')}") if err["xmlErrLevel"] == 1
                error("#{err.collect{|e|e.join(':')}.join('; ')}") if err["xmlErrLevel"] > 1
                # Display error message if any, but accept warnings and display whatever we got
                raise ("#{err.collect{|e|e.join(':')}.join('<br/>')}") if err["xmlErrLevel"] > 1
                # As an example: if you want to 'stay home' when serving a local copy of your favorite web site ;-)
                # Rewrite hardcoded absolute links to http://www.gentoo.org/ to /
                # res.body = out.gsub('href="http://www.gentoo.org/', 'href="/')
                res.body = out
              end
              res['Content-Type'] = if res.body =~ /^<\?xml/ then 'text/xml' else 'text/html' end
            end
            if mstat then
              res['ETag'] = makeETag(mstat)
              res['Last-Modified'] = mstat.mtime.httpdate
            end
          rescue => ex
            if ex.respond_to?(:errCode) then
              # One of ours (Gorg::Status::HTTPStatus)
              res.body = ex.html
              res.status = ex.errCode
              ex.header.each {|k,v| res[k]=v unless k =~ /status/i}
            else
              # Some ruby exceptions occurred, make it a syserr
              syserr = Gorg::Status::SysError.new
              res.body = syserr.html(ex)
              res.status = syserr.errCode
            end
          end
        end
      end
    end
  end
end

###
#|#  Start Here 
###

def www
  # Log accesses to either stderr, syslog or a file
  if $Config["accessLog"] == "syslog"
    # Use syslog again, use our own format based on default but without timestamp
    access_log = [ [ @syslog, "HIT   %h \"%r\" %s %b" ] ]
    STDERR.close
  elsif $Config["accessLog"] == "stderr"
    # Use syslog again, use our own format based on default but without timestamp
    access_log = [ [ STDERR, "HIT   %h \"%r\" %s %b" ] ]
  else
    # Open file and use it, if it's not writable, tough!
    access_log_stream = File.open($Config["accessLog"], "a")
    access_log = [ [ access_log_stream, WEBrick::AccessLog::COMBINED_LOG_FORMAT ] ]
    STDERR.close
  end

  s = WEBrick::HTTPServer.new( :AccessLog=>access_log, :Logger => $Log, :Port => $Config["port"])

  # Mount directories
  s.mount("/", GentooServlet, $Config["root"])

  # Start server
  trap("INT"){ s.shutdown }

  puts "\n\nStarting the Gorg web server on port #{$Config['port']}\n\nHit Ctrl-C or type \"kill #{$$}\" to stop it\n\n"

  s.start
end
